#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import os
import json
import textwrap
import itertools
from almdrlib.session import Session
from almdrlib.region import Region
from almdrlib.region import Residency
from almdrlib.docs.service import get_param_spec
from almdrlib.docs.service import format_json

def get_param_type(spec):
    type = spec.get('type')
    format = spec.get('format')

    if type == 'object':
        return 'list'
    elif type == 'array':
        return 'list'
    elif format:
        return format
    elif type:
        return type
    else:
        if 'oneOf' in spec:
            one_of_spec = spec['oneOf']
            return ' | '.join(list(set([
                    get_param_type(s)
                    for s in one_of_spec
                ])))
        elif 'anyOf' in spec:
            any_of_spec = spec['anyOf']
            return ' | '.join(list(set([
                    get_param_type(s)
                    for s in one_of_spec
                ])))
        else:
            raise ValueError(
                f'Unsupported parameter type. {json.dumps(spec, indent=2)}')

class FormatHelp:
    if sys.platform == 'win32':
        BOLD = ''
        UNDERLINE = ''
        END = ''
    elif hasattr(os, 'system') and os.system('(less) 2>/dev/null') == 0:
        BOLD = '\033[1m'
        UNDERLINE = '\033[4m'
        END = '\033[0m'
    else:
        BOLD = ''
        UNDERLINE = ''
        END = ''

    SECTION_BREAK = '\n'*2
    OPERATION_DESCIPTION_FOOTER = "See 'alcli help' for descriptions of global parameters."


class ALCliHelpFormatter():
    def __init__(self,
                width=80,
                indent_increment=2):
        self._name = ""
        self._width = width
        self._indent_increment = indent_increment
        self._initial_indent = '\t'
        self._subsequent_indent = '\t'

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, description):
        self._description = description


    def bold(self, msg):
        return f"{FormatHelp.BOLD}{msg}{FormatHelp.END}"

    def underline(self, msg):
        return f"{FormatHelp.UNDERLINE}{msg}{FormatHelp.END}"

    def wraptext(self, text, initial_indent=None, subsequent_indent=None):
        if text is None:
            return 

        if not initial_indent:
            initial_indent = self._initial_indent

        if not subsequent_indent:
            subsequent_indent = self._subsequent_indent

        for str in text.splitlines():
            yield textwrap.fill(
                        text=f"{str}",
                        width=self._width - 6, # Not sure why, but this seem to align description correctly
                        initial_indent=initial_indent,
                        subsequent_indent=subsequent_indent
                    )

    def make_header(self):
        title_name = self.name.upper() + "()"
        title = title_name + title_name.rjust(self._width - len(title_name))
        yield f'{title}'
        yield FormatHelp.SECTION_BREAK
        yield f'{self.bold("NAME")}'
        yield f'\t{self.name} - '
        yield ''

        
    def make_footer(self):
        footer_title = self.name.upper() + "()"
        fill_text = " " * (self._width - len(footer_title))
        yield f'{fill_text}{footer_title}'

    def make_description(self):
        if not self.description:
            yield ''
        yield f'{self.bold("DESCRIPTION")}'
        yield from self.wraptext(self.description)
        yield f''

class ALCliMainHelpFormatter(ALCliHelpFormatter):
    def __init__(self,
                services,
                width=80,
                indent_increment=2
                ):
        super().__init__(width, indent_increment)
        self.name = "alcli"
        self._services = services 
        self.description = (
                "The Alert Logic Command  Line  Interface is a tool "
                "to help you manage Alert Logic Services."
            )

    def make_synopsis(self):
        yield f'{self.bold("SYNOPSIS")}'
        yield f'\t  alcli [options] <command> <subcommand> [parameters]'
        synopsis = (
                f"\tUse {self.underline('alcli')} "
                f"{self.underline('command')} "
                f"{self.underline('help')} "
                "for information on a  specific command."
            )
        yield ''
        yield synopsis
        yield ''

    def make_options(self):
        yield f'{self.bold("OPTIONS")}'
        yield f'\t{self.bold("--debug")} (boolean)'
        yield ''
        yield '\tTurn on debug logging.'
        yield ''

        yield f'\t{self.bold("--query")} (string)'
        yield ''
        yield '\tA JMESPath query to use in filtering the response data.'
        yield ''

        yield f'\t{self.bold("--global_endpoint")} (string)'
        yield ''
        yield f'\tUse specific Alert Logic backend.'
        yield ''
        for endpoint in Region.list_endpoints():
            yield f'\to {endpoint}'
            yield ''

        yield f'\t{self.bold("--residency")} (string)'
        yield ''
        yield f'\tUse a specific data residency.'
        yield ''
        for residency in Residency.list_residencies():
            yield f'\to {residency}'
            yield ''

        yield f'\t{self.bold("--profile")} (string)'
        yield ''
        yield f'\tUse a specific profile from your configuration file.'
        yield ''

        yield f'\t{self.bold("--version")} (string)'
        yield ''
        yield '\tDisplay the version of alcli.'
        yield ''

    def make_services(self):
        yield f'{self.bold("AVAIABLE SERVICES")}'
        for service in self._services:
            yield f'\to {service}'
            yield ''

    def get_help(self):
        generators = []
        generators.append(self.make_header())
        generators.append(self.make_description())
        generators.append(self.make_synopsis())
        generators.append(self.make_options())
        generators.append(self.make_services())
        generators.append(self.make_footer())
        return iter(itertools.chain(*generators))


class ALCliServiceHelpFormatter(ALCliHelpFormatter):
    def __init__(self,
                name,
                spec,
                width=80,
                indent_increment=2
                ):
        super().__init__(width, indent_increment)
        self.name = name
        self.description = spec['info'].get('description', '')
        self._spec = spec 

    def make_operations(self):
        yield f'{self.bold("AVAIABLE COMMANDS")}'
        commands = sorted(self._spec.get('operations', {}).keys())
        for name in commands:
            yield f'\t o {name}'
            yield ''

    def get_help(self):
        generators = []
        generators.append(self.make_header())
        generators.append(self.make_description())
        generators.append(self.make_operations())
        generators.append(self.make_footer())
        return iter(itertools.chain(*generators))


class ALCliOperationHelpFormatter(ALCliHelpFormatter):
    def __init__(self,
                spec,
                width=80,
                indent_increment=2
                ):
        super().__init__(width, indent_increment)
        self.name = spec['operationId'] 
        self.description = spec['description']
        self._spec = spec 

        self._required_params = dict()
        self._params = dict()
        for key, value in spec.get('parameters', {}).items():
            if 'required' in value or 'x-alertlogic-required' in value:
                self._required_params[key] = value
            else:
                self._params[key] = value

    def make_description(self):
        if not self.description:
            yield ''
        yield f'{self.bold("DESCRIPTION")}'
        yield from self.wraptext(self.description)
        yield f''
        yield f"\tSee 'alcli help' for descriptions of global parameters."
        yield f''

    def make_synopsis(self):
        yield f'{self.bold("SYNOPSIS")}'
        yield f'\t  {self.name}'
        for name in self._required_params.keys():
            yield f'\t--{name} <value>'

        for name in self._params.keys():
            yield f'\t[--{name} <value>]'
        yield f''

    def make_options(self):
        yield f'{self.bold("OPTIONS")}'
        yield from self.make_parameters(self._required_params)
        yield from self.make_parameters(self._params)

    def make_output(self):
        yield f'{self.bold("OUTPUT")}'

        indent = '\t'
        spec = self._spec.get('response', {})
        if not bool(spec):
            yield f"{indent}This command doesn't produce an output"

        type = spec.get('type')
        if 'object' == type:
            yield f'{indent}{spec.get("title", "")} -> (structure)'
            indent = indent + '  '
            
            yield from self.wraptext(
                    spec.get('description'),
                    initial_indent=indent, subsequent_indent=indent)
            yield ''

            for name, spec in spec.get('properties', {}).items():
                yield f'{indent}{name} -> ({get_param_type(spec)})'
                yield from self.make_parameter(
                        spec, param_type='response', indent=indent + '  ', declare=True)
                yield ''
            return 

        elif 'array' == type:
            indent = indent + '  '
            yield f'{indent}{spec.get("x-alertlogic-name", "")} -> (list)'
        else:
            indent = indent + '  '

        yield from self.make_parameter(
                spec, param_type='response', indent=indent, declare=True)

    def make_parameters(self, params):
        for name, spec in params.items():
            if 'content' in spec:
                type = ' | '.join([
                        get_param_type(v)
                        for v in spec['content'].values()
                    ])
            else:
                type = get_param_type(spec)

            yield self.bold(f'\t--{name} ') + f'({type})'
            yield from self.make_parameter(
                    spec, param_type='request', indent='\t  ', declare=True)
            yield ''
            # print(f'{name}: {json.dumps(spec, indent=2)}')

    def make_parameter(self, spec, param_type='request', indent='\t', declare=False):
        type = spec.get('type')

        if type and declare:
        # if type:
            yield from self.wraptext(
                    spec.get('description'),
                    initial_indent=indent,
                    subsequent_indent=indent
                )
            yield ''

        if type == 'array':
            if 'items' in spec:
                yield from self.make_parameter(
                        spec['items'],
                        param_type=param_type, indent=indent, declare=declare)

        elif type == 'object':
            if declare and param_type == 'request':
                yield ''
                yield f'{indent}JSON Syntax:'
                yield ''
                json_indent=indent + '  '
                json_spec = format_json(
                        get_param_spec(spec),
                        json_indent).replace("'", "").replace("|", " | ")
                yield f'{json_indent}{json_spec}'
                yield ''

            properties = spec.get('properties', {})
            for prop_name, prop_spec in sorted(properties.items()):
                yield ''
                prop_title = (
                        f'o {self.bold(prop_name)} - '
                        f'{prop_spec.get("description", "")}'
                    )

                yield from self.wraptext(
                        prop_title,
                        initial_indent=indent,
                        subsequent_indent=indent + '  '
                    )

                yield from self.make_parameter(
                        spec=prop_spec,
                        param_type=param_type,
                        indent=indent + '  ',
                        declare=False
                    )

        elif 'oneOf' in spec:
            yield from self.make_indirect_parameter(
                'oneOf', spec.get('oneOf', []),
                param_type=param_type, indent=indent, declare=declare)

        elif 'anyOf' in spec:
                yield from self.make_indirect_parameter(
                    'anyOf', spec.get('anyOf', []),
                    param_type=param_type, indent=indent, declare=declare)
        elif 'content' in spec:
            yield from self.make_body_parameter(
                    spec.pop('content'), indent=indent, declare=declare)

        if 'enum' in spec:
            yield ''
            valid_values = ', '.join([f'{v}' for v in spec['enum']]) 
            yield f'{indent} Valid values: {valid_values}'

        if 'default' in spec:
            yield ''
            yield f'{indent} Default: {spec["default"]}'

    def make_indirect_parameter(self, indirect_type, spec, 
            param_type='request', indent='\t', declare=False):
        yield ''
        if indirect_type == 'oneOf':
            yield f'{indent}This argument must be one of the following:'
        else:
            yield f'{indent}This argument must be any of the following:'

        for s in spec:
            yield ''
            title = self.bold(s.get('title')) + \
                    f' ({get_param_type(s)}) - ' + \
                    s.get('description')
            yield from self.wraptext(
                    title,
                    initial_indent=indent,
                    subsequent_indent=indent + '  '
                )

            yield from self.make_parameter(
                    s, param_type, indent + '  ', declare)

    def make_body_parameter(self, spec, indent='\t', declare=False):
        yield f"{indent}This argument depends 'content_type' argument value"
        for content_type, content_spec in sorted(spec.items()):
            yield ''
            yield f'{indent}--content_type={content_type}:'
            yield from self.make_parameter(
                    spec=content_spec, indent=indent + '  ', declare=True)

    def get_help(self):
        generators = []
        generators.append(self.make_header())
        generators.append(self.make_description())
        generators.append(self.make_synopsis())
        generators.append(self.make_options())
        generators.append(self.make_output())
        generators.append(self.make_footer())
        return iter(itertools.chain(*generators))




import sys
import argparse
from difflib import get_close_matches
from almdrlib.client import OpenAPIKeyWord

HELP_MESSAGE = (
    "To see help text, you can run:\n"
    "\n"
    "  alcli help\n"
    "  alcli <command> help\n"
    "  alcli <command> <subcommand> help\n"
)
USAGE = (
    "alcli [options] <command> <subcommand> [<subcommand> ...] [parameters]\n"
    f"{HELP_MESSAGE}"
)


class ALCliParserUtils(object):
    OpenApiToNativeMap = {
        OpenAPIKeyWord.STRING: str,
        OpenAPIKeyWord.INTEGER: int,
        OpenAPIKeyWord.NUMBER: int,
        OpenAPIKeyWord.BOOLEAN: bool
    }

    @staticmethod
    def compound_schema(schema):
        compound_schema = schema.get(OpenAPIKeyWord.ANY_OF,
                                     schema.get(OpenAPIKeyWord.ONE_OF,
                                                schema.get(OpenAPIKeyWord.ALL_OF)))
        return compound_schema if isinstance(compound_schema, list) else False

    @staticmethod
    def detect_array_items_type(array_items_schema):
        compound_schema = ALCliParserUtils.compound_schema(array_items_schema)
        if compound_schema:
            type_maybe = compound_schema[0].get('type')
            if type_maybe:
                return type_maybe
            else:
                ALCliParserUtils.detect_array_items_type(compound_schema[0])

    @staticmethod
    def oapi_type_to_native(oapitype):
        return ALCliParserUtils.OpenApiToNativeMap.get(oapitype, str)


class CliHelpAction(argparse.Action):
    def __init__(self, option_strings, dest, formatter, **kwargs):
        print(f"CliHelpAction:__init__ called. option_strings: {option_strings}, dest: {dest}, formatter: {formatter}, kwargs: {kwargs}")
        super().__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        print(f"CliHelpAction:__call__ called: Namespace: {namespace}, Values: {values}, Options: {option_string}")
        setattr(namespace, self.dest, values)


class CliArgParserBase(argparse.ArgumentParser):
    # Number of choices per line
    ChoicesPerLine = 2

    def _check_value(self, action, value):
        # converted value must be one of the choices (if specified)
        if isinstance(action.choices, dict):
            choices =  list(action.choices.keys())
        else:
            choices = action.choices

        if choices is not None and value not in choices:
            if 'help' in choices:
                choices.remove('help')
            msg = ['Invalid choice, valid choices are:\n']
            for i in range(len(choices))[::self.ChoicesPerLine]:
                current = []
                for choice in choices[i:i+self.ChoicesPerLine]:
                    current.append('%-40s' % choice)
                msg.append(' | '.join(current))
            possible = get_close_matches(value, choices, cutoff=0.8)
            if possible:
                extra = ['\n\nInvalid choice: %r, maybe you meant:\n' % value]
                for word in possible:
                    extra.append('  * %s' % word)
                msg.extend(extra)
            raise argparse.ArgumentError(action, '\n'.join(msg))

class ALCliArgsParser(CliArgParserBase):
    """
        Main CLI Arguments Parser 
    """

    def __init__(self, services, version, description, prog=None):
        super().__init__(
                formatter_class=argparse.RawTextHelpFormatter,
                add_help=False,
                conflict_handler='resolve',
                description=description,
                usage=USAGE,
                prog=prog)

        self.add_argument('--version', action="version",
                          version=version,
                          help='Display the version of this tool')

        # self.add_argument('help', action=CliHelpAction, formatter="1234")
        # self.add_argument('help')
        self._subparsers = self.add_subparsers(
                dest='service',
                title='service',
                parser_class=ServicesArgsParser,
                required=False)
        self._create_parsers(services)

    def parse_known_args(self, args=None, namespace=None):
        parsed_args, remaining = super().parse_known_args(args, namespace)
        # print(f"ALCliArgsParser:parse_known_args returning {parsed_args}, {remaining}")
        return parsed_args, remaining

    #
    # For all know services, create a subparser
    #
    def _create_parsers(self, services):
        self._subparsers.add_parser('help', service=None)
        for name, service in services.items():
            self._subparsers.add_parser(name, service=service)

class ServicesArgsParser(CliArgParserBase):
    """
        Alert Logic Services Parser 
    """

    def __init__(self, *args, **kwargs):
        self._service = kwargs.pop('service')
        super().__init__(
                formatter_class = argparse.RawTextHelpFormatter,
                add_help=False,
                conflict_handler='resolve',
                usage=USAGE)
        
    def parse_known_args(self, args=None, namespace=None):
        #
        # Get Service API Schema
        #
        if self._service is None:
            return super().parse_known_args(args, namespace)

        service_api = self._service.get_service_api(self._service.name)

        self._required=True
        subparsers = self.add_subparsers(
                dest='operation',
                title="operation",
                required=True,
                parser_class=OperationArgsParser)

        #
        # Add subparsers for all service operations
        #
        for op_name, op_spec in service_api['operations'].items():
            operation_parser = subparsers.add_parser(op_name, spec=op_spec)
        
        #
        # Add help command
        #
        subparsers.add_parser('help', spec=None)
        return super().parse_known_args(args, namespace)


class OperationArgsParser(CliArgParserBase):
    """
        Alert Logic Service Operation Parser
    """

    def __init__(self, *args, **kwargs):
        self._spec = kwargs.pop('spec')
        super().__init__(
            formatter_class=argparse.RawTextHelpFormatter,
            add_help=False,
            conflict_handler='resolve',
            usage=USAGE)

    def make_parameter_argument(self, schema):
        type = schema.get(OpenAPIKeyWord.TYPE)
        required = schema.get(OpenAPIKeyWord.REQUIRED, False)
        parser_type = ALCliParserUtils.oapi_type_to_native(schema.get('type'))
        if type == 'array':
            items_type = ALCliParserUtils.detect_array_items_type(schema.get('items', {}))
            if items_type in OpenAPIKeyWord.SIMPLE_DATA_TYPES:
                return {'nargs': '+', 'type': ALCliParserUtils.oapi_type_to_native(items_type), 'required': required}
            else:
                return {'required': required}
        else:
            return {'required': required, 'type': parser_type}

    def parse_known_args(self, args=None, namespace=None):
        # Add Operation arguments to the parser
        if 'help' in args:
            self.add_argument('help')
            return super().parse_known_args(args, namespace)
        elif self._spec is not None:
            for name, schema in self._spec[OpenAPIKeyWord.PARAMETERS].items():
                kwargs = self.make_parameter_argument(schema)
                type = schema.get(OpenAPIKeyWord.TYPE)
                if type == 'boolean':
                    self.add_argument(f"--{name}",
                                      action='store_true')
                else:
                    self.add_argument(f"--{name}", **kwargs)
        return super().parse_known_args(args, namespace)



class CLIArgument(object):
    def __init__(self, name):
        self._name = name

    @property
    def cli_name(self):
        if self._positional_arg:
            return self._name
        else:
            return '--' + self._name

    def add_to_parser(self, parser):
        cli_name = self.cli_name

    def add_to_parser(self, parser):
        kwargs = {}
        parser.add_argument(cli_name, **kwargs)

    @property
    def name(self):
        return self._name



import re
import ast
import sys
from urllib import request
import json

from setuptools import setup, find_packages
if sys.platform == 'win32':
    from cx_Freeze import setup, Executable
    base = None
    executables = [Executable(
            script='alcli/alertlogic_cli.py',
            targetName='alcli',
            base=base,
            icon='icons/alertlogic-win.ico'
            )
        ]
else:
    executables = []

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

try:
    # This is to force definitions to be upgraded every build not related to the definitions change
    sdk_pypi_url = "https://pypi.org/pypi/alertlogic-sdk-python/json"
    with request.urlopen(sdk_pypi_url) as defs_rq:
        defs_info = json.loads(defs_rq.read())
    sdk_latest_version = defs_info['info']['version']
    sdk_dependency = 'alertlogic-sdk-python>=' + sdk_latest_version
except:
    sdk_dependency = 'alertlogic-sdk-python>=v1.0.47'

requirements = [
        sdk_dependency,
        'configparser==4.0.2',
        'pyyaml==5.1.2',
        'jmespath==0.9.4',
        'importlib_metadata==1.6.0'
    ]
setup(
    name='alcli',
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    url='https://github.com/alertlogic/alcli',
    license='MIT license',
    author='Alert Logic Inc.',
    author_email='devsupport@alertlogic.com',
    python_requires='>=3.7',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'
    ],
    description='The Alert Logic Command Line Utility (CLI).',
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/markdown',
    entry_points = {
        'console_scripts': [
            'alcli = alcli.alertlogic_cli:main'
        ]
    },
    scripts=[],
    packages=find_packages(exclude=['contrib', 'docs', 'tests*', 'troubleshooting']),
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=requirements,
    extras_require={
        'dev': [
            'pytest>=3',
            'mock>=2.0.0',
            'httpretty>=0.8.14',
            'pycodestyle>=2.3.1'
        ],
    },
    options = {
        'build_exe': {
            'packages': ['os', 'sys', 'ctypes', 'jsonschema', 'almdrlib', 'alcli'],
            'excludes': ['tkinter','tcl','ttk'],
            'include_msvcr': True
        },
        'bdist_msi': {
            'add_to_path': False
        }
    },
    executables = executables,
    keywords=['alcli', 'almdr', 'alertlogic', 'alertlogic-cli']
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import os
import platform
import json
import jmespath
from json import JSONDecodeError
import logging
import argparse
import re
import pydoc

from urllib.parse import urlparse
from collections import OrderedDict
from pydoc import pager


import almdrlib
from almdrlib.session import Session
from almdrlib.client import Client
from almdrlib.client import OpenAPIKeyWord
from almdrlib.region import Region
from almdrlib.region import Residency
from almdrlib.version import version as almdrlib_version

from alcli.cliparser import ALCliArgsParser
from alcli.cliparser import ALCliParserUtils
from alcli.cliparser import USAGE
from alcli.clihelp import ALCliMainHelpFormatter
from alcli.clihelp import ALCliServiceHelpFormatter
from alcli.clihelp import ALCliOperationHelpFormatter
from alcli.version import version as alcli_version

from alsdkdefs.version import version as alsdkdefs_version

if getattr(sys, 'frozen', False):
    # frozen
    dir_ = os.path.dirname(sys.executable)
else:
    # unfrozen
    dir_ = os.path.dirname(os.path.dirname(__file__))

sys.path.insert(0, dir_)
# sys.path.insert(0, os.path.dirname(__file__))

logger = logging.getLogger('alcli.almdr_cli')
LOG_FORMAT = (
    '%(asctime)s - %(threadName)s - %(name)s - %(levelname)s - %(message)s')

GLOBAL_ARGUMENTS = [
        'access_key_id',
        'secret_key',
        'profile',
        'residency',
        'global_endpoint',
        'query',
        'debug'
    ]

def cli_pager(text):
    """The first time this is called, determine what kind of pager to use."""
    global cli_pager
    cli_pager = get_cli_pager()
    cli_pager(text)

def get_cli_pager():
    if sys.platform == 'win32':
        return lambda text: pydoc.pipepager(text, 'more /C')
    if hasattr(os, 'system') and os.system('(less) 2>/dev/null') == 0:
        return lambda text: pydoc.pipepager(text, 'less -R')

class AlertLogicCLI(object):
    def __init__(self):
        self._subparsers = None
        self._services = None
        self._arguments = None

    def main(self, args=None):
        args = args or sys.argv[1:]
        services = self._get_services()
        parser = self._create_parser(services)

        parsed_args, remaining = parser.parse_known_args(args)
        logger.debug(f"Parsed Arguments: {parsed_args}, Remaining: {remaining}")
        
        if parsed_args.service == 'help' or parsed_args.service is None:
            AlertLogicCLI.show_help(
                     ALCliMainHelpFormatter(Session.list_services())
                )
            return 128

        if parsed_args.operation == 'help':
            AlertLogicCLI.show_help(
                    ALCliServiceHelpFormatter(
                        parsed_args.service,
                        Session.get_service_api(parsed_args.service)
                    )
                )
            return 128

        if hasattr(parsed_args, 'help') and \
                hasattr(parsed_args, 'service') and \
                hasattr(parsed_args, 'operation') and \
                parsed_args.help == 'help':
            spec = Session.get_service_api(parsed_args.service)['operations']
            AlertLogicCLI.show_help(
                    ALCliOperationHelpFormatter(
                        spec.get(parsed_args.operation, {})
                    )
                )
            return 0

        try:
            return services[parsed_args.service](remaining, parsed_args)
        except almdrlib.exceptions.AlmdrlibValueError as e:
            sys.stderr.write(f"{e}\n")
            return 255
        except almdrlib.session.AuthenticationException as e:
            sys.stderr.write("Access Denied\n")
            return 255
        except Exception as e:
            template = "An exception of type {0} occurred. Arguments:\n{1!r}"
            message = template.format(type(e).__name__, e.args)
            logger.exception(message)
            return 255

    def _get_services(self):
        if self._services is None:
            self._services = OrderedDict()
            services_list = Session.list_services()
            self._services = {
                service_name: ServiceOperation(name=service_name)
                for service_name in services_list
            }

        return self._services

    def _create_parser(self, services):
        parser = ALCliArgsParser(
                services,
                f"alcli/{alcli_version} Python/{platform.python_version()}"
                f" almdrlib/{almdrlib_version}"
                f" alsdkdefs/{alsdkdefs_version}",
                "Alert Logic CLI Utility",
                prog="alcli")
        
        # Add Global Options
        parser.add_argument('--access_key_id', dest='access_key_id', default=None)
        parser.add_argument('--secret_key', dest='secret_key', default=None)
        parser.add_argument('--profile', dest='profile', default=None)
        parser.add_argument('--residency', dest='residency', default=None, choices=Residency.list_residencies())
        parser.add_argument('--global_endpoint', dest='global_endpoint', default=None, choices=Region.list_endpoints())
        parser.add_argument('--query', dest='query', default=None)
        parser.add_argument('--debug', dest='debug', default=False, action="store_true")
        return parser

    @staticmethod
    def show_help(help_generator):
        # cli_pager(help_formatter.format_page() + '\n')
        cli_pager('\n'.join(list(help_generator.get_help())) + '\n')
        return 0


class ServiceOperation(object):
    """
        A service operation. For example: alcli aetuner would create
        a ServiceOperation object for aetuner service
    """
    def __init__(self, name):
        self._name = name
        self._client= None
        #self._service = None
        self._description = None
        self._operations = None

    def __call__(self, args, parsed_globals):
        operation_name = ""
        kwargs = {}
        for name, value in parsed_globals.__dict__.items():
            if name == 'operation':
                operation_name = value
            elif name == 'debug': 
                if value:
                    almdrlib.set_logger('almdrlib', logging.DEBUG, format_string=LOG_FORMAT)
            elif name == 'service':
                continue
            elif name in GLOBAL_ARGUMENTS:
                continue
            else:
                kwargs[name] = value
        
        service = self._init_service(parsed_globals)
        operation = service.operations.get(operation_name, None)
        if operation:
            # Remove optional arguments that haven't been supplied
            op_args = {k:self._encode(operation, k, v) for (k,v) in kwargs.items() if v is not None}
            res = operation(**op_args)
            try:
                self._print_result(res.json(), parsed_globals.query)
            except json.decoder.JSONDecodeError:
                print(f'HTTP Status Code: {res.status_code}')

    def get_service_api(self, service_name):
        return Session.get_service_api(service_name=service_name)

    @property
    def client(self):
        if self._client is None:
            self._client = Client(self._name)
        return self._client

    @property
    def name(self):
        return self._name

    @property
    def description(self):
        if self._description is None:
            self._description = self.client.description
        return self._description

    @property
    def operations(self):
        if self._operations is None:
            self._operations = self.client.operations
        return self._operations

    def _init_service(self, parsed_globals):
        account_id = None
        if hasattr(parsed_globals, 'account_id'):
            account_id = parsed_globals.account_id

        return almdrlib.client(
                self._name,
                access_key_id=parsed_globals.access_key_id,
                secret_key=parsed_globals.secret_key,
                account_id=account_id,
                profile=parsed_globals.profile,
                global_endpoint=parsed_globals.global_endpoint,
                residency=\
                        parsed_globals.residency and \
                        parsed_globals.residency or "default"
        )

    def _encode(self, operation, param_name, param_value):
        if isinstance(param_value, str):
            p = urlparse(param_value)
            if p.scheme == "file":
                value_file_path = os.path.abspath(os.path.join(p.netloc, p.path))
                with open(value_file_path, "r") as value_file:
                    param_value = value_file.read()

        schema = operation.get_schema()
        parameter = schema[OpenAPIKeyWord.PARAMETERS][param_name]
        parameter_type = parameter.get(OpenAPIKeyWord.TYPE)
        oneof = parameter.get(OpenAPIKeyWord.ONE_OF)
        anyof = parameter.get(OpenAPIKeyWord.ANY_OF)
        allof = parameter.get(OpenAPIKeyWord.ALL_OF)

        if oneof or anyof or allof:
            # Attempt to find type in the decomposed schema
            find_type = [t.get('type') for t in allof or anyof or oneof if t.get('type')]
            parameter_type = find_type[0] if find_type else None

        if not parameter_type:
            # If failed to figure parameter_type out, fallback to object
            parameter_type = OpenAPIKeyWord.OBJECT

        if parameter_type in OpenAPIKeyWord.SIMPLE_DATA_TYPES:
            return param_value

        if parameter_type == OpenAPIKeyWord.ARRAY:
            items_type = ALCliParserUtils.detect_array_items_type(schema.get('items', {}))
            if items_type in OpenAPIKeyWord.SIMPLE_DATA_TYPES:
                return param_value
            else:
                try:
                    return json.loads(param_value)
                except JSONDecodeError as e:
                    logging.error(f"{e}")

        if parameter_type in OpenAPIKeyWord.OBJECT:
            try:
                return json.loads(param_value)
            except JSONDecodeError as e:
                #
                # this could be an argument encoded using a short form such as
                # key1=value,key2=value2
                #
                regex = re.compile(
                        '(?P<pair>(?P<key>.+?)(?:=)(?P<value>[^=]+)(?:,|$))')
                result = dict(
                    (m.groupdict()['key'], m.groupdict()['value'])
                    for m in regex.finditer(param_value)
                )
                return result

        # TODO Raise an exception if we can't build a dictionary from the provided input
        return param_value

    def _print_result(self, result, query):
        if query:
            result = jmespath.search(query, result) 
        print(f"{json.dumps(result, sort_keys=True, indent=4)}")


def main():
    if os.environ.get("DEBUG"):
        print("Running in DEBUG mode")
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.ERROR)

    cli= AlertLogicCLI()
    cli.main()

if __name__ == "__main__":
    main()




import sys
import argparse
from difflib import get_close_matches

HELP_MESSAGE = (
    "To see help text, you can run:\n"
    "\n"
    "  alcli help\n"
    "  alcli <command> help\n"
    "  alcli <command> <subcommand> help\n"
)
USAGE = (
    "alcli [options] <command> <subcommand> [<subcommand> ...] [parameters]\n"
    f"{HELP_MESSAGE}"
)

class CommandAction(argparse.Action):
    def __init__(self, option_strings, dest, command_table, **kwargs):
        self.command_table = command_table
        super(CommandAction, self).__init__(
            option_strings, dest, choices=self.choices, **kwargs
        )
         
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, values)

    @property
    def choices(self):
        return list(self.command_table.keys())

    @choices.setter
    def choices(self, val):
        # argparse.Action will always try to set this value upon
        # instantiation, but this value should be dynamically
        # generated from the command table keys. So make this a
        # NOOP if argparse.Action tries to set this value.
        pass
         
class CLIArgParser(argparse.ArgumentParser):
    Formatter = argparse.RawTextHelpFormatter

    # Number of choices per line
    ChoicesPerLine = 2

    def _check_value(self, action, value):
        # converted value must be one of the choices (if specified)
        if action.choices is not None and value not in action.choices:
            msg = ['Invalid choice, valid choices are:\n']
            for i in range(len(action.choices))[::self.ChoicesPerLine]:
                current = []
                for choice in action.choices[i:i+self.ChoicesPerLine]:
                    current.append('%-40s' % choice)
                msg.append(' | '.join(current))
            possible = get_close_matches(value, action.choices, cutoff=0.8)
            if possible:
                extra = ['\n\nInvalid choice: %r, maybe you meant:\n' % value]
                for word in possible:
                    extra.append('  * %s' % word)
                msg.extend(extra)
            raise argparse.ArgumentError(action, '\n'.join(msg))
    
    def __init__(self, command_table, version_string,
                    description, argument_table, prog=None):
        super(CLIArgParser, self).__init__(
            formatter_class=self.Formatter,
            add_help=False,
            conflict_handler='resolve',
            description=description,
            usage=USAGE,
            prog=prog)
        self._build(command_table, version_string, argument_table)

    def parse_known_args(self, args, namespace=None):
            parsed, remaining = super(CLIArgParser, self).parse_known_args(args, namespace)
            return parsed, remaining
        
    def _build(self, command_table, version_string, argument_table):
        self.add_argument('--version', action="version",
                          version=version_string,
                          help='Display the version of this tool')
        self.add_argument('--profile', action="store",
                          help="Specify cnfiguration profile name to use")
        self.add_argument('command', action=CommandAction,
                          command_table=command_table)

